from .abstract import CutoffSettings


class MarkCutoffSettings(CutoffSettings):
    """Класс настроек, при которых для нахождения отсечек сравнивается разметка локальных конфигураций"""

    def __init__(self):
        self.min_by_mark = {}

    def check_cutoff(self, event, order_settings):
        config = order_settings.config(event)
        m = config.mark()

        return m in self.min_by_mark and order_settings.cmp_events(self.min_by_mark[m], event, config2=config) < 0, {
            "mark": m}

    def update(self, event, order_settings, *, mark=None, **kwargs):
        if mark is None:
            mark = order_settings(event)

        self.min_by_mark[mark] = event



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to, add_place, add_transition


def generate_dining_philosophers(n, add_host=False):
    """
    Генерирует сеть Петри, моделирующую задачу об обедающих философах
    :param n: количество философов
    :param add_host: если True, генерируется версия с хостом
    :return: сеть Петри и её начальную разметку
    """
    net = PetriNet(f"Dining philosophers {n}")

    def place(x):
        return add_place(net, x)

    def trans(x, label):
        return add_transition(net, x, label)

    def arc(f, t):
        return add_arc_from_to(f, t, net)

    marking = Marking()

    chopsticks = [place(f"chopstick{x}") for x in range(1, n + 1)]
    marking.update(chopsticks)

    if add_host:
        host = place("host")
        marking[host] = n - 1

    for i in range(1, n + 1):
        chopstick1 = chopsticks[i - 1]
        chopstick2 = chopsticks[i % n]

        thinking = place(f"thinking{i}")
        prepare_left = place(f"prep l{i}")
        prepare_right = place(f"prep r{i}")
        ready_left = place(f"ready l{i}")
        ready_right = place(f"ready r{i}")
        dining = place(f"dining{i}")

        marking[thinking] += 1

        prepare = trans(name := f"prep{i}", label=name)
        take_left = trans(name := f"take l{i}", label=name)
        take_right = trans(name := f"take r{i}", label=name)
        eat = trans(name := f"eat{i}", label=name)
        think = trans(name := f"think{i}", label=name)

        arc(thinking, prepare)
        if add_host:
            arc(host, prepare)

        arc(prepare, prepare_left)
        arc(prepare, prepare_right)

        arc(prepare_left, take_left)
        arc(chopstick1, take_left)
        arc(take_left, ready_left)

        arc(prepare_right, take_right)
        arc(chopstick2, take_right)
        arc(take_right, ready_right)

        arc(ready_left, eat)
        arc(ready_right, eat)
        arc(eat, dining)

        arc(dining, think)
        arc(think, chopstick1)
        arc(think, chopstick2)
        arc(think, thinking)

        if add_host:
            arc(think, host)

    return net, marking



from abc import ABC, abstractmethod

from pm4py.objects.petri_net.obj import Marking


class OrderSettings(ABC):
    """Интерфейс настроек"""

    @abstractmethod
    def config(self, event):
        """
        Генерирует локальную конфигурацию события
        :param event: событие
        :return: локальную конфигурацию event
        """

    @abstractmethod
    def cmp_events(self, e1, e2, *, config1=None, config2=None, **kwargs):
        """
        Сравнивает события
        :param e1: первое событие
        :param e2: второе событие
        :param config1: локальная конфигурация первого события. Используется для оптимизации
        :param config2: локальная конфигурация второго события. Используется для оптимизации
        :param kwargs: "подсказки", ускоряющие вычисления. Каждый класс настроек может принимать различные "подсказки"
        :return: отрицательное число, если e1 < e1; положительное число, если e1 > e2; 0 в других случаях
        """


# Скопированная реализация collections.Counter._keep_positive (это внутренний метод, который может исчезнуть
# в последующих версиях python)
def _keep_positive(counter):
    nonpositive = [elem for elem, count in counter.items() if not count > 0]
    for elem in nonpositive:
        del counter[elem]
    return counter


class Configuration(ABC):
    """Абстрактный класс конфигурации"""

    @abstractmethod
    def __iter__(self):
        """Перечисляет все события в конфигурации"""

    @abstractmethod
    def __len__(self):
        """Количество событий в конфигурации"""

    def mark(self):
        """
        Вычисляет разметку среза конфигурации
        :return: нужную разметку
        """
        res = Marking()
        for e in self:
            post_set_marking = e.postset_marking()
            pre_set_marking = e.preset_marking()

            res.update(post_set_marking)
            res.subtract(pre_set_marking)

        return _keep_positive(res)



"""
В этом модуле реализованы основные алгоритмы, необходимые для построения префикса развертки:
1. Очередь с приоритетом для хранения и быстрого извлечения возможных расширений (priority_queue.py)
2. Хранение и обновление отношения co (concurrency_relation.py)
3. Поиск возможных расширений (possible_extensions.py)
"""
from .concurrency_relation import *
from .possible_extensions import *
from .priority_queue import *

__all__ = [
    "Co",
    "update_possible_extensions",
    "PriorityQueue"
]



"""
В этом модуле представлены реализации декораций - классов, отвечающих за параметры вывода построенных префиксов
(параметр decorations у pm4py.view_petri_net). Декорации конструируются во время построения префикса.
Все встроенные классы декораций (кроме Idle) устроены по принципу декораторов, таким образом декорации можно
комбинировать.

В модуле представлены следующие декорации:
1. idle - пустые декорации, используется как основание для остальных декораций (в цепочке декораторов)
2. labels - декорации, позволяющие настроить надписи на элементах префикса
3. colors - декорации, позволяющие настроить цвет элементов префикса
4. abstract - интерфейс декораций для новых реализаций
"""

from .abstract import Decorations
from .colors import ColorsDecorations
from .idle import IdleDecorations
from .labels import LabelsDecorations

__all__ = [
    "Decorations",
    "IdleDecorations",
    "LabelsDecorations",
    "ColorsDecorations"
]



from itertools import chain

from pm4py.objects.petri_net.obj import Marking, PetriNet
from pm4py.objects.petri_net.utils import petri_utils

from .event import NSafeEvent


def update_possible_extensions(pe, new, c, co):
    """
    Обновление очереди возможных расширений после добавления нового события
    :param pe: очередь возможных расширений
    :param new: добавленное событие
    :param c: множество условий, конкурентных с new (считается в Co.update)
    :param co: отношение конкурентности
    """
    transitions = set(
        chain.from_iterable(
            petri_utils.post_set(x) | petri_utils.pre_set(x)
            for x in (y.place for y in petri_utils.post_set(new))
        )
    )

    new_postset = petri_utils.post_set(new)
    for t in transitions:
        preset = petri_utils.pre_set(t)
        postset = petri_utils.post_set(t)
        n = len(preset | postset)
        preset_weight = {x.source: x.weight for x in t.in_arcs}
        preset_conds = [x for x in new_postset if
                        x.place in postset and x.place not in preset or x.place in preset and x.markers >=
                        preset_weight[x.place]]
        cover(pe, t, preset_conds, co, c, preset, postset, preset_weight, n)


def cover(pe, t, preset_conds, co, c, preset, postset, preset_weight, n):
    if len(preset_conds) == n:
        e = NSafeEvent(t, Marking({x.place: x.markers for x in preset_conds}))
        for c in preset_conds:
            arc = PetriNet.Arc(c, e)
            e.in_arcs.add(arc)
        pe.add((e, set(preset_conds)))
    else:
        left = preset - set(x.place for x in preset_conds)
        if left:
            p = next(iter(left))
            choose_from = (x for x in c if x.place == p and x.markers >= preset_weight[p])
        else:
            left = postset - set(x.place for x in preset_conds)
            p = next(iter(left))
            choose_from = (x for x in c if x.place == p)

        for d in choose_from:
            new_c = [x for x in c if co(d, x) and x != d]
            preset_conds.append(d)
            cover(pe, t, preset_conds, co, new_c, preset, postset, preset_weight, n)
            preset_conds.pop()



from collections import defaultdict

from .abstract import Decorations


class IdleDecorations(Decorations):
    """Пустые декорации"""

    def __init__(self):
        self.decorations = defaultdict(lambda: {})

    def get(self):
        return self.decorations



from .abstract import OrderSettings


class IdleOrderSettings(OrderSettings):
    """Настройки порядка, при которой все конфигурации считаются несравнимыми. Таким образом, очередь с приоритетом pe
     будет работать как обычный список, т.е. возможные расширения будут рассматриваться в произвольном порядке.
     Такой порядок не является адекватным - он используется для построения развертки "до бесконечности", т.е. пока
     развертка не кончится или не будет достигнут предел количества событий в префиксе.
     """

    def config(self, e):
        return None

    def cmp_events(self, *args, **kwargs):
        return 0



from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils


class Condition(PetriNet.Place):
    """
    Класс, представляющий условие развертки
    :attribute place: позиция изначальной сети, которой помечено условие
    """

    def __init__(self, place):
        super().__init__("")  # Название он получит при добавлении в префикс (см. obj/prefix.py)
        self.place = place  # Позиция в сети, соответствующая данному условию

    # Единственное событие в preset-е
    @property
    def input_event(self):
        e, = petri_utils.pre_set(self)  # По определению развертки, preset условия состоит из одного события
        return e

    def __repr__(self):
        return f"<{self.name} of {self.place.name}>"

    @property
    def net_label(self):
        return id(self.place)



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to as arc


def generate_mutual_exclusion(n, m):
    """
    Генерирует сеть Петри, моделирующую семафор
    :param n: количество работающих потоков
    :param m: максимальное значение счетчика семафора
    :return: сеть Петри и её начальную разметку
    """
    net = PetriNet(f"Mutual exclusion {n},{m}")
    marking = Marking()

    semaphore = PetriNet.Place("semaphore")
    net.places.add(semaphore)
    marking.update((semaphore,) * m)

    for i in range(1, n + 1):
        critical = PetriNet.Place(f"critical{i}")
        isolated = PetriNet.Place(f"isolated{i}")
        waiting = PetriNet.Place(f"waiting{i}")

        net.places.update((critical, isolated, waiting))
        marking.update((isolated,))

        capture = PetriNet.Transition(name := f"capture{i}", label=name)
        release = PetriNet.Transition(name := f"release{i}", label=name)
        await_ = PetriNet.Transition(name := f"await{i}", label=name)

        net.transitions.update((capture, release, await_))

        arc(semaphore, capture, net)
        arc(waiting, capture, net)
        arc(capture, critical, net)

        arc(critical, release, net)
        arc(release, semaphore, net)
        arc(release, isolated, net)

        arc(isolated, await_, net)
        arc(await_, waiting, net)

    return net, marking



from itertools import product

from pm4py.objects.petri_net.utils import petri_utils


class Co:
    """
    Класс, представляющий отношение co(x, y) := [not (x <= y) and not (x # y)] на условиях сети
    """

    def __init__(self):
        self.pairs = set()  # Отношение представляется в виде множества пар

    def __call__(self, f, s):
        """Определяет, конкурентны ли условия"""
        return (f, s) in self.pairs or (s, f) in self.pairs

    def update(self, new_event, conditions):
        """
        Обновляет отношение после добавления нового события
        :param new_event: новое событие
        :param conditions: множество всех условий в префиксе (в том числе новых)
        """
        new = petri_utils.post_set(new_event)  # post_set нового события
        old = conditions - new  # Все остальные условия

        concurrent_with = []  # Старые условия, с которыми конкурентны все условия из preset-а нового события
        # (а значит и само новое событие)

        for c in old:
            for c1 in petri_utils.pre_set(new_event):
                if not self(c, c1) or c == c1:
                    break
            else:  # Если `c` конкурентно со всеми условиями из preset-а нового события
                concurrent_with.append(c)

        self.pairs.update((x, y) for x, y in product(new, new) if x != y)
        self.pairs.update((x, y) for x, y in product(new, concurrent_with))

        return concurrent_with

    def __repr__(self):
        return repr(self.pairs)



"""
В этом модуле представлены реализации настроек порядка - классов, задача которых - генерировать локальные конфигурации
событий и определять адекватный порядок на конфигурациях.

В модуле представлены следующие настройки:
1. Idle - все конфигурации считаются несравнимыми
2. basic - конфигурации представляются обычным множеством и сравниваются по длинам
(базовый алгоритм МакМиллана)
3. foata - конфигурации представляются списком "слоев", соответствующих Фоатовской нормальной форме, и сравниваются
по отсортированным меткам (TotalOrderSettings) или по Фоатовской нормальной форме (FoataOrderSettigs)
4. abstract - интерфейс конфигураций и настроек порядка для новых реализаций

Все встроенные настройки порядка (кроме idle) включаются сравнение длин конфигураций, поэтому длины конфигураций
при использовании этих настроек событий кэшируются
"""

from .abstract import OrderSettings, Configuration
from .basic import BasicOrderSettings
from .foata import FoataOrderSettings
from .idle import IdleOrderSettings

__all__ = [
    "BasicOrderSettings",
    "FoataOrderSettings",
    "OrderSettings",
    "Configuration",
    "IdleOrderSettings"
]



from collections import defaultdict
from itertools import chain, product, combinations

from pm4py.objects.petri_net.obj import Marking, PetriNet
from pm4py.objects.petri_net.utils import petri_utils

from ..obj import Event


def check_coset(s, co):
    for x, y in combinations(s, 2):
        if not co(x, y):
            return False
    return True


def update_possible_extensions(pe, new, c, co):
    """
    Обновление очереди возможных расширений после добавления нового события
    :param pe: очередь возможных расширений
    :param new: добавленное событие
    :param c: множество условий, конкурентных с new (считается в Co.update)
    :param co: отношение конкурентности
    """
    transitions = set(
        chain.from_iterable(
            petri_utils.post_set(x)
            for x in set(y.place for y in petri_utils.post_set(new))
        )
    )

    new_postset = petri_utils.post_set(new)
    for t in transitions:
        n = sum(x.weight for x in t.in_arcs)
        preset = petri_utils.pre_set(t)
        preset_marking = Marking(x.source for x in t.in_arcs for _ in range(x.weight))

        preset_by_place = defaultdict(lambda: [])
        for x in new_postset:
            if x.place in preset:
                preset_by_place[x.place].append(x)
        for cond in c:
            if cond.place in preset_by_place and cond not in new_postset:  # Второе условие нужно только при добавлении \bot
                preset_by_place[cond.place].append(cond)

        for choice in product(
                *(list(combinations(preset_by_place[cond], preset_marking[cond])) for cond in preset_by_place)):
            preset_conds = list(chain.from_iterable(choice))

            for cond in preset_conds:
                if cond in new_postset:
                    break
            else:  # Если в выбранном coset-е нет новых элементов
                continue  # то это расширение уже добавлено (если нужно)
            if not check_coset(preset_conds, co):
                continue

            cover(pe, t, preset_conds, co, c, preset_marking, n)

        # preset_conds = list(filter(lambda x: x.place in preset, petri_utils.post_set(new)))
        # cover(pe, t, preset_conds, co, c, preset_marking, n)


def cover(pe, t, preset_conds, co, c, preset_marking, n):
    if len(preset_conds) == n:
        e = Event(t)
        for c in preset_conds:
            arc = PetriNet.Arc(c, e)
            e.in_arcs.add(arc)
        pe.add((e, set(preset_conds)))
    else:
        left = set(preset_marking) - set(x.place for x in preset_conds)
        p = next(iter(left))

        possible_additions = [x for x in c if x.place == p]
        for choice in combinations(possible_additions, preset_marking[p]):
            if not check_coset(choice, co):
                continue
            new_c = [x for x in c if x not in choice and all(co(d, x) for d in choice)]
            preset_conds.extend(choice)
            cover(pe, t, preset_conds, co, new_c, preset_marking, n)

            for _ in range(preset_marking[p]):
                preset_conds.pop()

        # for d in filter(lambda x: x.place == p, c):
        #     new_c = list(filter(lambda x: co(d, x) and x != d, c))
        #     preset_conds.append(d)
        #     cover(pe, t, preset_conds, co, new_c, preset_marking, n)
        #     preset_conds.pop()



from .abstract import Decorations
from .idle import IdleDecorations


class LabelsDecorations(Decorations):
    """Декорации, позволяющие изменять надписи на элементах сети"""

    def __init__(self, inner=None, /, *, conditions, events):
        """
        :param inner: внутренняя декорация, по умолчанию idle
        :param conditions: функция, возвращающая надпись на условии
        :param events: функция, возвращающая надпись на событии
        """
        self.events = events
        self.inner = inner or IdleDecorations()
        self.conditions = conditions

    @classmethod
    def standard(cls, inner=None, /):
        """
        Базовые декорации для стандартного алгоритма. Надписи условий и событий будут совпадать с именами позиций и
        переходов, которыми они помечены
        :param inner: внутренняя декорация, по умолчанию idle
        :return:
        """
        return LabelsDecorations(inner,
                                 conditions=lambda c: c.place.name,
                                 events=lambda e: e.transition.name if e.transition is not None else ""
                                 )

    @classmethod
    def n_safe(cls, inner=None, /):
        """
        Базовые декорации для n-safe алгоритма. Надписи условий будут иметь вид "{name}:{markers}", где name -
        имя позиции начальной сети, которой помечено условие, markers - количество маркеров в этой позиции, которому
        соответствует условию.
        Аналогично, надписи событий будут иметь вид: "{name}"
        :param inner: внутренняя декорация (по умолчанию idle)
        :return:
        """
        return LabelsDecorations(inner,
                                 conditions=lambda c: f"{c.place.name}:{c.markers}",
                                 events=lambda e: f"{e.transition.name}"
                                 )

    def add_condition(self, c):
        self.inner.add_condition(c)
        self.inner.get()[c]["label"] = self.conditions(c)

    def add_event(self, e):
        self.inner.add_event(e)
        self.inner.get()[e]["label"] = self.events(e)

    def get(self):
        return self.inner.get()



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to, add_place, add_transition


def generate_slotted_ring(n):
    net = PetriNet(f"Milner's scheduler {n}")

    def place(x):
        return add_place(net, x)

    def trans(x):
        return add_transition(net, x, x)

    def arc(f, t):
        return add_arc_from_to(f, t, net)

    marking = Marking()

    p1s = [place(f"p1_{x}") for x in range(1, n + 1)]
    p2s = [place(f"p2_{x}") for x in range(1, n + 1)]
    p10s = [place(f"p10_{x}") for x in range(1, n + 1)]

    marking.update(p10s)

    for i in range(1, n + 1):
        p1 = p1s[i - 1]
        p2 = p2s[i - 1]
        p10 = p10s[i - 1]

        np1 = p1s[i % n]
        np2 = p2s[i % n]
        pp10 = p10s[(i - 2) % n]

        p3, p4, p5, p6, p7, p8, p9 = (place(f"p{x}_{i}") for x in range(3, 10))
        marking[p7] += 1

        owner = trans(f"owner{i}")
        other = trans(f"other{i}")
        go_on = trans(f"go on{i}")
        write = trans(f"write{i}")
        give_free_slot = trans(f"gfs{i}")
        put_message_in_slot = trans(f"put{i}")
        int_ack = trans(f"int ack{i}")
        free = trans(f"free{i}")
        used = trans(f"used{i}")
        ack = trans(f"ack{i}")

        arc(p1, owner)
        arc(p1, other)

        arc(owner, p2)
        arc(p2, go_on)
        arc(p2, write)

        arc(go_on, p3)
        arc(p3, give_free_slot)

        arc(give_free_slot, p4)
        arc(put_message_in_slot, p4)
        arc(p4, ack)

        arc(other, p5)
        arc(write, p5)
        arc(p5, put_message_in_slot)

        arc(int_ack, p6)
        arc(p6, give_free_slot)
        arc(p6, put_message_in_slot)

        arc(give_free_slot, p7)
        arc(p7, free)

        arc(put_message_in_slot, p8)
        arc(p8, used)

        arc(free, p9)
        arc(used, p9)
        arc(p9, int_ack)

        arc(p10, free)
        arc(p10, used)

        arc(ack, pp10)
        arc(used, np1)
        arc(free, np2)

    return net, marking



"""
В этом модуле представлены генераторы различных сетей Петри, а именно:
1. mutual_exclusion - модель семафора с заданным количеством работников и флагов
2. dining_philosophers - модель задачи об обедающих философах (заданного их количества)
с возможной взаимной блокировкой.
"""

from .dining_philosophers import generate_dining_philosophers
from .mutual_exclusion import generate_mutual_exclusion
from .dining_philosophers_with_dict import generate_dining_philosophers_with_dict
from .milners_cyclic_scheduler import generate_milners_cyclic_scheduler
from .generate_slotted_ring import generate_slotted_ring

__all__ = [
    "generate_mutual_exclusion",
    "generate_dining_philosophers",
    "generate_dining_philosophers_with_dict",
    "generate_milners_cyclic_scheduler",
    "generate_slotted_ring"
]



"""
В этом модуле реализованы классы, расширяющие представления сети Петри библиотеки pm4py:
1. Event расширяет PetriNet.Transition и представляет событие развертки
2. Condition расширяет PetriNet.Place и представляет условие разверкти
3. Prefix расширяет PetriNet и представляет префикс развертки
"""

from .condition import *
from .event import *
from .prefix import *

__all__ = [
    "Condition",
    "Event",
    "Prefix"
]



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to, add_place, add_transition


def generate_milners_cyclic_scheduler(n, round_cap=None):
    """
    Генерирует сеть Петри модели циклического планировщика Мильнера
    :param n: количество рабочих
    :param round_cap: если передано, в сеть добавляется позиция, ограничивающая количество раундов
    :return: сеть Петри, начальная разметка
    """
    net = PetriNet(f"Milner's scheduler {n}")

    def place(x):
        return add_place(net, x)

    def trans(x):
        return add_transition(net, x, x)

    def arc(f, t):
        return add_arc_from_to(f, t, net)

    marking = Marking()

    ready_st = [place(f"ready st{x}") for x in range(1, n + 1)]
    ready_nr = [place(f"ready nr{x}") for x in range(1, n + 1)]

    ready_new = place("ready_new")
    marking[ready_new] += 1

    new_round = trans("new round")
    arc(ready_new, new_round)
    arc(new_round, ready_st[0])
    arc(new_round, ready_nr[0])

    if round_cap is not None:
        rounds = place("rounds")
        marking[rounds] = round_cap
        arc(rounds, new_round)

    for i in range(1, n + 1):
        my_ready_st = ready_st[i - 1]
        my_ready_nr = ready_nr[i - 1]

        next_ready_st = None if i == n else ready_st[i]
        next_ready_nr = ready_new if i == n else ready_nr[i]

        ready = place(f"ready{i}")
        working = place(f"working{i}")
        finished = place(f"finished{i}")

        marking[ready] += 1

        work = trans(f"work{i}")
        finish = trans(f"finish{i}")
        next_round = trans(f"next{i}")

        arc(my_ready_st, work)
        arc(ready, work)
        arc(work, working)
        if next_ready_st is not None:
            arc(work, next_ready_st)

        arc(working, finish)
        arc(finish, finished)

        arc(finished, next_round)
        arc(my_ready_nr, next_round)
        arc(next_round, ready)
        arc(next_round, next_ready_nr)

    return net, marking



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to, add_place, add_transition


def generate_dining_philosophers_with_dict(n):
    """
    Генерирует сеть Петри модели обедающих философов со словарем
    :param n: количество философов
    :return: сеть Петри, начальная разметка
    """
    net = PetriNet(f"Dining philosophers with dict {n}")

    def place(x):
        return add_place(net, x)

    def trans(x, label):
        return add_transition(net, x, label)

    def arc(f, t):
        return add_arc_from_to(f, t, net)

    marking = Marking()

    chopsticks = [place(f"chopstick{x}") for x in range(1, n + 1)]
    marking.update(chopsticks)

    dicts = [place(f"dict{x}") for x in range(1, n + 1)]
    no_dicts = [place(f"no dict{x}") for x in range(1, n + 1)]

    marking[dicts[0]] += 1
    marking.update(no_dicts[1:])

    for i in range(1, n + 1):
        chopstick1 = chopsticks[i - 1]
        chopstick2 = chopsticks[i % n]

        my_no_dict = no_dicts[i - 1]
        my_dict = dicts[i - 1]

        next_no_dict = no_dicts[i % n]
        next_dict = dicts[i % n]

        thinking = place(f"thinking{i}")
        prepare_left = place(f"prep l{i}")
        prepare_right = place(f"prep r{i}")
        ready_left = place(f"ready l{i}")
        ready_right = place(f"ready r{i}")
        dining = place(f"dining{i}")

        marking[thinking] += 1

        prepare = trans(name := f"prep{i}", label=name)
        take_left = trans(name := f"take l{i}", label=name)
        take_right = trans(name := f"take r{i}", label=name)
        eat = trans(name := f"eat{i}", label=name)
        think = trans(name := f"think{i}", label=name)
        pass_dict = trans(name := f"pass{i}", name)

        arc(thinking, prepare)
        arc(my_no_dict, prepare)
        arc(prepare, prepare_left)
        arc(prepare, prepare_right)

        arc(prepare_left, take_left)
        arc(chopstick1, take_left)
        arc(take_left, ready_left)

        arc(prepare_right, take_right)
        arc(chopstick2, take_right)
        arc(take_right, ready_right)

        arc(ready_left, eat)
        arc(ready_right, eat)
        arc(eat, dining)

        arc(dining, think)
        arc(think, chopstick1)
        arc(think, chopstick2)
        arc(think, thinking)
        arc(think, my_no_dict)

        arc(my_dict, pass_dict)
        arc(next_no_dict, pass_dict)
        arc(pass_dict, my_no_dict)
        arc(pass_dict, next_dict)

    return net, marking



from .abstract import CutoffSettings


class IdleCutoffSettings(CutoffSettings):
    r"""Класс настроек, при которых никакое событие не будет отсечено. Это соответствует C_e = \emptyset"""

    def check_cutoff(self, event, order_settings):
        return False, {}

    def update(self, event, order_settings, *, mark=None, **kwargs):
        pass



from pm4py.objects.petri_net.obj import Marking
from pm4py.objects.petri_net.utils import petri_utils

from ...obj import Event


# Отличие от обычных событий в том, что теперь они помечены не только переходом в изначальной сети,
# но и разметкой preset-а
class NSafeEvent(Event):
    def __init__(self, transition, marking):
        super().__init__(transition)
        self.marking_hash = hash(marking)

    def __repr__(self):
        t_name = r"\bot" if self.transition is None else self.transition.name
        return f"<{self.name} of {t_name} with hash {self.marking_hash}>"

    @property
    def net_label(self):
        return super().net_label, self.marking_hash

    def postset_marking(self):
        return Marking({c.place: c.markers for c in petri_utils.post_set(self)})

    def preset_marking(self):
        return Marking({c.place: c.markers for c in petri_utils.pre_set(self)})



"""
В этом модуле представлены реализации алгоритмов построения префиксов развертки, а именно:
1. standard - стандартный алгоритм построения канонических префиксов
2. n_safe - построение канонических префиксов развертки сети, трансформированной в безопасную.
"""

from .n_safe import build_prefix as n_safe_algorithm
from .standard import build_prefix as standard_algorithm

__all__ = [
    "standard_algorithm",
    "n_safe_algorithm",
]



class ConfigLength:
    """Класс для кэширования длин локальных конфигураций событий"""

    def __init__(self, config_type):
        self.cache = {}
        self.conf = config_type

    def calculated(self, ev):
        return ev in self.cache

    def save(self, ev, length):
        self.cache[ev] = length

    def __call__(self, ev):
        res = self.cache.get(ev, len(self.conf(ev)))
        self.save(ev, res)
        return res



from pm4py.objects.petri_net.utils import petri_utils

from .condition import NSafeCondition
from .event import NSafeEvent
from .possible_extensions import update_possible_extensions
from ...alg import PriorityQueue, Co
from ...decorations import IdleDecorations
from ...obj import Prefix


def build_prefix(net, m0, order_settings, cutoff_settings, decorations=None, event_count=None,
                 pe_optimise_local_config=True, order_of_adding=None):
    """
    Строит канонический префикс развертки данной сети (при правильно определенном контексте срезания, за который
    отвечают аргументы order_settings и cutoff_settings)
    :param net: сеть Петри
    :param m0: начальная разметка net
    :param order_settings: настройки порядка
    :param cutoff_settings: настройки срезания
    :param decorations: декорации (опционально)
    :param event_count: максимальное количество событий в префиксе. Если построение префикса не закончится до того,
    как количество событий в префиксе превысит данное значение, построение префикса прекратится и значение флага
    finished будет False. Если не передать event_count, ограничения на количество событий не будет
    :param pe_optimise_local_config: аргумент для инициализации PriorityQueue
    :param order_of_adding: список, в который будут сохранены события в порядке добавления к префиксу (опционально)
    :return: префикс развертки
    """
    if order_of_adding is None:
        order_of_adding = []

    if event_count is not None and event_count <= 0:
        raise ValueError("event count must be positive")
    if decorations is None:
        decorations = IdleDecorations()
    res = Prefix(net.name)

    e = NSafeEvent(None, None)
    bot = e
    res.add_event(e)

    co = Co()
    pe = PriorityQueue(order_settings, pe_optimise_local_config)

    # Начальные условия берутся из всех позиций изначальной сети
    for p in net.places:
        c = NSafeCondition(p, m0.get(p, 0))
        res.add_condition(c)
        petri_utils.add_arc_from_to(e, c, res)
        decorations.add_condition(c)
        decorations.add_starting_condition(c)

    co.update(e, res.places)
    cutoff_settings.update(e, order_settings, mark=m0)
    update_possible_extensions(pe, e, res.places, co)

    count = 0
    finished = True
    while pe:
        e, pre = pe.pop()
        res.add_event(e)
        order_of_adding.append(e)

        preset_weight = {x.source: x.weight for x in e.transition.in_arcs}
        postset_weight = {x.target: x.weight for x in e.transition.out_arcs}

        for c in pre:
            petri_utils.add_arc_from_to(c, e, res)
            p = c.place
            # Единственное отличие от основного алгоритма - подсчёт количества маркеров в новых позициях
            dm = c.markers - preset_weight.get(p, 0) + postset_weight.get(p, 0)
            c = NSafeCondition(p, dm)
            res.add_condition(c)
            petri_utils.add_arc_from_to(e, c, res)
            decorations.add_condition(c)

        is_cutoff, hint = cutoff_settings.check_cutoff(e, order_settings)
        decorations.add_event(e)
        if not is_cutoff:
            concurrent_with = co.update(e, res.places)
            cutoff_settings.update(e, order_settings, **hint)
            update_possible_extensions(pe, e, concurrent_with, co)
            count += 1
            if event_count is not None and pe and count >= event_count:
                finished = False
                break
        else:
            res.add_cutoff(e)
            decorations.add_cutoff_event(e)

    for a in bot.out_arcs:
        a.target.in_arcs.remove(a)
        res.arcs.remove(a)

    res.transitions.remove(bot)
    res.finished = finished

    return res



from abc import ABC, abstractmethod


class CutoffSettings(ABC):
    """Интерфейс настроек отсечения"""

    @abstractmethod
    def check_cutoff(self, event, order_settings):
        """
        Проверка, является ли событие отсечкой
        :param event: проверяемое событие
        :param order_settings: настройки порядка
        :return: кортеж (res, hint), где res - результат проверки (True/False), hint - словарь "подсказок" - значений,
        которые будут переданы как аргументы в CutoffSettings.update. Подсказки нужны для оптимизации - зачастую в
        check_cutoff и update производятся схожие вычисления
        """

    @abstractmethod
    def update(self, event, order_settings, **kwargs):
        """
        Обновление объекта настроек при добавлении нового события
        :param event: новое событие
        :param order_settings: настройки порядка
        :param kwargs: "подсказки" для ускорения вычислений. Сюда передаются подсказки из метода check_cutoff
        """



from ...obj import Condition


# Отличие от обычных условий в том, что теперь они помечены не только позицией из начальной сети,
# но и количеством маркеров в этой позиции
class NSafeCondition(Condition):
    def __init__(self, place, markers):
        super().__init__(place)
        self.markers = markers

    def __repr__(self):
        return f"<{self.name} of {self.place.name} with {self.markers} markers>"

    @property
    def net_label(self):
        return super().net_label, self.markers



from pm4py.objects.petri_net.obj import PetriNet


# Класс, представляющий префикс развёртки
class Prefix(PetriNet):
    """
    Класс, представляющий префикс развертки
    :attribute finished: закончилось ли построения префикса или нет (нет в случае, если при построении было поставлено
    ограничение на количество событий, которое было превышено)
    :attribute cutoff_events: множество событий-отсечек
    """

    def __init__(self, net_name):
        super().__init__(f"Finite complete prefix of '{net_name}'")
        # Счётчики событий и условий для наименования
        self.event_counter = 0
        self.condition_counter = 1
        self.cutoff_events = set()
        self.finished = False

    def add_condition(self, c):
        c.name = f"c{self.condition_counter}"
        self.condition_counter += 1
        self.places.add(c)

    def add_event(self, e):
        e.name = f"e{self.event_counter}"
        self.event_counter += 1
        self.transitions.add(e)

    def add_cutoff(self, ev):
        self.cutoff_events.add(ev)



# Реализация очереди с приоритетом.
# Используется для нахождения возможного расширения с наименьшей локальной конфигурацией.
# В очереди будут лежать пары (event, preset), где event - возможное расширение, preset - условия, которые
# будут составлять его preset.
class PriorityQueue:
    """
    Реализация очереди с приоритетом.
    Используется для нахождения возможного расширения с наименьшей локальной конфигурацией.
    В очереди будут лежать пары (event, preset), где event - возможное расширение, preset - условия, которые
    будут составлять его preset.
    """

    def __init__(self, order_settings, optimise_local_config=True):
        """
        :param order_settings: настройки порядка
        :param optimise_local_config: определяет, считать локальные конфигурации сравниваемых событий
        заранее и передавать их как подсказки. Ускоряет работу алгоритма, если настройки порядка ускоряются при передаче
        подсказок config1 и config22
        """
        self.cmp = order_settings.cmp_events
        self.conf = order_settings.config
        self.heap = []
        self.optimise_local_config = optimise_local_config

    @staticmethod
    def _left_child(ind):
        return 2 * ind + 1

    @staticmethod
    def _right_child(ind):
        return 2 * ind + 2

    @staticmethod
    def _parent(ind):
        return (ind - 1) // 2

    def _in_bounds(self, ind):
        return 0 <= ind < len(self.heap)

    def _heapify_down(self, ind):
        if not self._in_bounds(ind):
            return

        cv_config = self.conf(self.heap[ind][0]) if self.optimise_local_config else None
        lv_config = None
        rv_config = None
        while True:
            li = self._left_child(ind)
            ri = self._right_child(ind)

            lv = self.heap[li][0] if self._in_bounds(li) else None
            rv = self.heap[ri][0] if self._in_bounds(ri) else None
            cv = self.heap[ind][0]

            if lv is not None:
                lv_config = self.conf(lv)
            if rv is not None:
                rv_config = self.conf(rv)

            if lv is not None and self.cmp(cv, lv, config1=cv_config, config2=lv_config) > 0 and (
                    rv is None or self.cmp(rv, lv, config1=rv_config, config2=lv_config) >= 0):
                self.heap[ind], self.heap[li] = self.heap[li], self.heap[ind]
                ind = li
                cv_config = lv_config
            elif rv is not None and self.cmp(cv, rv, config1=cv_config, config2=rv_config) > 0 and (
                    lv is None or self.cmp(lv, rv, config1=lv_config, config2=rv_config) >= 0):
                self.heap[ind], self.heap[ri] = self.heap[ri], self.heap[ind]
                ind = ri
                cv_config = rv_config
            else:
                break

    def _heapify_up(self, ind):
        if not self._in_bounds(ind):
            return
        while True:
            pi = self._parent(ind)
            if not self._in_bounds(pi):
                break

            pv = self.heap[pi][0]
            cv = self.heap[ind][0]

            if self.cmp(cv, pv) < 0:
                self.heap[ind], self.heap[pi] = self.heap[pi], self.heap[ind]
                ind = pi
            else:
                break

    def add(self, el):  # Добавить новое потенциальное событие
        ni = len(self.heap)
        self.heap.append(el)
        self._heapify_up(ni)

    def pop(self):  # Извлечь потенциальное событие с минимальной локальной конфигурацией
        self.heap[0], self.heap[-1] = self.heap[-1], self.heap[0]

        r = self.heap.pop()
        self._heapify_down(0)
        return r

    def __len__(self):
        return len(self.heap)

    def __repr__(self):
        return repr(self.heap)



r"""
В это м модуле представлены реализации настроек отсечения - классов, задача которых - выбирать, с каким конфигурациями
развертки будет сравниваться локальная конфигурация потенциального события-отсечки (это соответствует \approx и
\{C_e\}_{e \in E} в контексте срезания)

В модуле представлены следующие настройки:
1. idle - никакие события не будут отсечками
2. mark - соответствует \approx(c, c') := Mark(c) = Mark(c') и C_e = C_{loc}
3. abstract - интерфейс настроек для новых реализаций
"""
from .abstract import CutoffSettings
from .idle import IdleCutoffSettings
from .mark import MarkCutoffSettings

__all__ = [
    "CutoffSettings",
    "MarkCutoffSettings",
    "IdleCutoffSettings"
]



from pm4py.objects.petri_net.utils import petri_utils

from ..alg import Co, PriorityQueue, update_possible_extensions
from ..decorations import IdleDecorations
from ..obj import Prefix, Event, Condition


def build_prefix(net, m0, order_settings, cutoff_settings, decorations=None, event_count=None,
                 pe_optimise_local_config=True, order_of_adding=None):
    """
    Строит канонический префикс развертки данной сети (при правильно определенном контексте срезания, за который
    отвечают аргументы order_settings и cutoff_settings)
    :param net: сеть Петри
    :param m0: начальная разметка net
    :param order_settings: настройки порядка
    :param cutoff_settings: настройки срезания
    :param decorations: декорации (опционально)
    :param event_count: максимальное количество событий в префиксе. Если построение префикса не закончится до того,
    как количество событий в префиксе превысит данное значение, построение префикса прекратится и значение флага
    finished будет False. Если не передать event_count, ограничения на количество событий не будет
    :param pe_optimise_local_config: аргумент для инициализации PriorityQueue
    :param order_of_adding: список, в который будут сохранены события в порядке добавления к префиксу (опционально)
    :return: префикс развертки
    """
    if order_of_adding is None:
        order_of_adding = []

    if event_count is not None and event_count <= 0:
        raise ValueError("event count must be positive")
    if decorations is None:
        decorations = IdleDecorations()

    res = Prefix(net.name)  # В res будет находиться итоговый префикс

    e = Event(None)  # "Изначальное" событие \bot. Его post_set-ом будут условия, соответствующе начальной разметке
    bot = e
    res.add_event(e)

    co = Co()
    pe = PriorityQueue(order_settings, optimise_local_config=pe_optimise_local_config)

    for p in m0.elements():  # Добавление условий, соответствующих начальной разметке
        c = Condition(p)
        res.add_condition(c)
        petri_utils.add_arc_from_to(e, c, res)
        decorations.add_condition(c)
        decorations.add_starting_condition(c)

    # Обновление отношения co, очереди pe и настроек срезания
    co.update(e, res.places)
    cutoff_settings.update(e, order_settings, mark=m0)
    update_possible_extensions(pe, e, res.places, co)

    count = 0
    finished = True
    while pe:  # Пока к префиксу можно добавить новые события
        e, pre = pe.pop()  # Извлекаем минимальное событие

        # Добавляем к префиксу выбранное событие и дуги к нему от условий в его preset-е
        res.add_event(e)
        order_of_adding.append(e)
        for c in pre:
            petri_utils.add_arc_from_to(c, e, res)

        # Создаём условия, соответствующие позициям из post_set-а перехода, представленного выбранным событием
        for a in e.transition.out_arcs:
            for _ in range(a.weight):
                c = Condition(a.target)
                res.add_condition(c)
                petri_utils.add_arc_from_to(e, c, res)
                decorations.add_condition(c)

        # Проверяем, является ли событие отсечкой
        is_cutoff, hint = cutoff_settings.check_cutoff(e, order_settings)
        decorations.add_event(e)
        if not is_cutoff:
            # Если нет, обновляем co и pe
            concurrent_with = co.update(e, res.places)
            cutoff_settings.update(e, order_settings, **hint)
            update_possible_extensions(pe, e, concurrent_with, co)
            count += 1
            if event_count is not None and pe and count >= event_count:  # Проверяем, что не превышен предел
                # количества событий
                finished = False
                break
        else:
            # Иначе добавляем в cutoff_events
            res.add_cutoff(e)
            decorations.add_cutoff_event(e)

    # После процедуры удаляем событие bot из префикса
    for a in bot.out_arcs:
        a.target.in_arcs.remove(a)
        res.arcs.remove(a)

    res.transitions.remove(bot)

    res.finished = finished

    return res



from functools import partial
from itertools import chain, zip_longest

from pm4py.objects.petri_net.utils import petri_utils

from .abstract import Configuration, OrderSettings
from .config_length_utils import ConfigLength


class FoataConfiguration(Configuration):
    """Представление конфигурации в виде "слоев", соответствующих Фоатовской нормальной форме"""

    def __init__(self, event, save_length=None):
        self.events = []
        met = set()
        self.init_events(event, met)

        if save_length is not None:
            save_length.save(event, len(met))

    def init_events(self, event, met, before=0):
        if event in met:
            return
        met.add(event)

        if before == 0:
            self.events.insert(0, [event])
            before += 1
        else:
            self.events[before - 1].append(event)

        for c in petri_utils.pre_set(event):
            self.init_events(c.input_event, met, before - 1)

    def __iter__(self):
        return chain.from_iterable(self.events)

    def __len__(self):
        return sum(1 for _ in self)

    def __repr__(self):
        return repr(self.events)


def _cmp_from_operators(o1, o2):
    if o1 < o2:
        return -1
    if o1 > o2:
        return 1
    return 0


def lex_order(events):
    return sorted(events, key=lambda e: e.net_label)


def compare_lex(seq1, seq2):
    for f, s in zip_longest(seq1, seq2):
        if f is None:  # seq1 - префикс seq2
            return -1
        if s is None:  # seq2 - префикс seq1
            return 1

        # fi = id(f.transition)
        # si = id(s.transition)

        cmp = _cmp_from_operators(f.net_label, s.net_label)
        if cmp != 0:
            return cmp

    return 0


def cmp_events(e1, e2, config_length, *, config1=None, config2=None, **kwargs):
    if config1 is not None and not config_length.calculated(e1):
        config_length.save(e1, len(config1))
    if config2 is not None and not config_length.calculated(e2):
        config_length.save(e2, len(config2))

    # Конфигурации не вычисляются, если уже посчитана длина - возможно, сравнения длин будет достаточно
    c1 = config1 if config_length.calculated(e1) else FoataConfiguration(e1, config_length)
    c2 = config2 if config_length.calculated(e2) else FoataConfiguration(e2, config_length)

    # Сравнение длин
    l1 = config_length(e1)
    l2 = config_length(e2)

    l_cmp = l1 - l2
    if l_cmp != 0:
        return l_cmp

    # Считаем не посчитанные конфигурации
    if c1 is None:
        c1 = FoataConfiguration(e1, config_length)

    if c2 is None:
        c2 = FoataConfiguration(e2, config_length)

    # Сравнение в лексикографическом порядке
    total_cmp = compare_lex(lex_order(c1), lex_order(c2))

    if total_cmp != 0:
        return total_cmp

    # Сравнение нормальных форм
    for f, s in zip_longest(c1.events, c2.events):
        if f is None:  # seq1 - префикс seq2
            return -1
        if s is None:  # seq2 - префикс seq1
            return 1

        cmp = compare_lex(lex_order(f), lex_order(s))

        if cmp != 0:
            return cmp

    return 0


class FoataOrderSettings(OrderSettings):
    """Настройки порядка, при которых конфигурации сравниваются по Фоатовской нормальной форме"""

    def __init__(self):
        config_length = ConfigLength(FoataConfiguration)
        self.conf = partial(FoataConfiguration, save_length=config_length)
        self.cmp = partial(cmp_events, config_length=config_length)

    def config(self, event):
        return self.conf(event)

    def cmp_events(self, e1, e2, **kwargs):
        """
        Принимаемые подсказки:
            config1 - конфигурация e1. Если длина конфигурации e1 не была посчитана жо этого, вместо вычисления
            конфигурации будет использована длина config1. Если длины e1 и e2 равны, вместо вычисления конфигурации
            будет использована config1.
            config2 - конфигурация e2. Используется аналогично config1
        """
        return self.cmp(e1, e2, **kwargs)



from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open("requirements.txt") as f:
    requires = f.read().split("\n")

setup(
    name="pm4py-pn-unfoldings",
    version="1.0.2",
    author="TimurTimergalin",
    author_email="tmtimergalin8080@gmail.com",
    description="Library for Petri net unfoldings based on pm4py",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/TimurTimergalin/Unfoldings",
    packages=find_packages(include="unfoldings*"),
    install_requires=requires,
    classifiers=[
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    keywords=["petri-nets unfoldings pm4py"]
)



from .abstract import Decorations
from .idle import IdleDecorations


class ColorsDecorations(Decorations):
    """Декорации, позволяющие менять цвет элементов сети"""

    def __init__(self, inner=None, /, *, conditions="#ffffff", events="#ffffff", starting_conditions=None, cutoff_events=None):
        """
        :param inner: внутренняя декорация, по умолчанию idle
        :param conditions: цвет условий, по умолчанию на выбор pm4py
        :param events: цвет событий, по умолчанию на выбор pm4py
        :param starting_conditions: цвет начальных событий, по умолчанию совпадает с conditions
        :param cutoff_events: цвет отсеченных событий, по умолчанию совпадает с events
        """
        self.events = events
        self.starting_conditions = starting_conditions
        self.cutoff_events = cutoff_events
        self.conditions = conditions
        self.inner = inner or IdleDecorations()

    def add_event(self, e):
        self.inner.add_event(e)
        if self.events is not None:
            self.inner.get()[e]["color"] = self.events

    def add_cutoff_event(self, e):
        self.inner.add_cutoff_event(e)
        if self.cutoff_events is not None:
            self.inner.get()[e]["color"] = self.cutoff_events

    def add_condition(self, c):
        self.inner.add_condition(c)
        if self.conditions is not None:
            self.inner.get()[c]["color"] = self.conditions

    def add_starting_condition(self, c):
        self.inner.add_starting_condition(c)
        if self.starting_conditions is not None:
            self.inner.get()[c]["color"] = self.starting_conditions

    def get(self):
        return self.inner.get()



from abc import ABC, abstractmethod


class Decorations(ABC):
    """Интерфейс декораций"""

    def add_event(self, e):
        """
        Вызывается, когда событие (отсеченное или нет) добавляется в сеть
        :param e: добавленное событие
        """

    def add_cutoff_event(self, e):
        """
        Вызывается, когда событие помечается как событие-отсечка. Этот метод вызывается после add_event
        :param e: событие-отсечка
        """

    def add_condition(self, c):
        """
        Вызывается, когда условие добавляется в сеть
        :param c: добавленное условие
        """

    def add_starting_condition(self, c):
        """
        Вызывается, когда в сеть добавляют начальное условие. Этот метод вызывается после add_condition
        :param c: стартовое условие
        """

    @abstractmethod
    def get(self):
        """
        Метод для получения словаря полученных декораций, который затем можно передать в pm4py.view_petri_net
        :return: словарь декораций
        """
        pass



from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils import petri_utils


class Event(PetriNet.Transition):
    """
    Класс, представляющий событие развертки
    :attribute transition: переход изначальной сети, которым помечено событие
    """

    def __init__(self, transition):  # None, если event = \bot
        # Название он получит при добавлении в префикс (см. obj/prefix.py)
        super().__init__("", "")
        self.transition = transition  # Переход в сети, соответствующий данному событию

    def __repr__(self):
        bot = r'\bot'
        return f"<{self.name} of {self.transition.name if self.transition else bot}>"

    @property
    def net_label(self):
        return id(self.transition)

    def postset_marking(self):
        return Marking(c.place for c in petri_utils.post_set(self))

    def preset_marking(self):
        return Marking(c.place for c in petri_utils.pre_set(self))



from functools import partial

from pm4py.objects.petri_net.utils import petri_utils

from .abstract import Configuration, OrderSettings
from .config_length_utils import ConfigLength


class BasicConfiguration(Configuration):
    """Представление конфигурации в виде множества событий"""

    def __init__(self, event, save_length=None):
        self.events = set()
        self.init_events(event)

        if save_length is not None:
            save_length.save(event, len(self.events))

    def init_events(self, event):
        if event in self.events:
            return
        self.events.add(event)
        for c in petri_utils.pre_set(event):
            self.init_events(c.input_event)

    def __iter__(self):
        return iter(self.events)

    def __len__(self):
        return len(self.events)

    @staticmethod
    def cmp_events(e1, e2, config_length, *, config1=None, config2=None, **kwargs):
        if config1 is not None and not config_length.calculated(e1):
            config_length.save(e1, len(config1))
        if config2 is not None and not config_length.calculated(e2):
            config_length.save(e2, len(config2))

        l1 = config_length(e1)
        l2 = config_length(e2)
        return l1 - l2

    def __repr__(self):
        return repr(self.events)


class BasicOrderSettings(OrderSettings):
    """Настройки порядка, при которых конфигурации сравниваются по длине"""

    def __init__(self):
        config_length = ConfigLength(BasicConfiguration)
        self.conf = partial(BasicConfiguration, save_length=config_length)
        self.cmp = partial(BasicConfiguration.cmp_events, config_length=config_length)

    def config(self, event):
        return self.conf(event)

    def cmp_events(self, e1, e2, **kwargs):
        """
        Принимаемые подсказки:
            config1 - конфигурация e1. Если длина конфигурации e1 не была посчитана жо этого, вместо вычисления
            конфигурации будет использована длина config1
            config2 - конфигурация e2. Используется аналогично config1
        """
        return self.cmp(e1, e2, **kwargs)
